#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library

# import interfaces
from pyams_security.interfaces import IRole
from zope.schema.interfaces import IVocabularyFactory

# import packages
from pyams_utils.request import check_request
from zope.interface import implementer, provider
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import getVocabularyRegistry, SimpleVocabulary, SimpleTerm


@implementer(IRole)
class Role(object):
    """Role utility class"""

    id = FieldProperty(IRole['id'])
    title = FieldProperty(IRole['title'])
    description = FieldProperty(IRole['description'])
    permissions = FieldProperty(IRole['permissions'])
    managers = FieldProperty(IRole['managers'])

    def __init__(self, values=None, **args):
        if not isinstance(values, dict):
            values = args
        self.id = values.get('id')
        self.title = values.get('title')
        self.description = values.get('description')
        self.permissions = values.get('permissions')
        self.managers = values.get('managers')


def register_role(config, role):
    """Register a new role

    Roles registry is not required.
    But only registered roles can be applied via default
    ZMI features.

    If a role is registered several times, previous registrations
    will just be updated to add new permissions.
    Title and description are not updated after first registration.
    """
    registry = config.registry
    if not IRole.providedBy(role):
        role_utility = registry.queryUtility(IRole, name=role.get('id'))
        if role_utility is None:
            # registering a new role
            role_utility = Role(role)
            registry.registerUtility(role_utility, IRole, name=role_utility.id)
        else:
            # new registration of a given role to add permissions
            role_utility.permissions = (role_utility.permissions or set()) | \
                                       (role.get('permissions') or set())
            role_utility.managers = (role_utility.managers or set()) | \
                                    (role.get('managers') or set())
    else:
        registry.registerUtility(role, IRole, name=role.id)


@provider(IVocabularyFactory)
class RolesVocabulary(SimpleVocabulary):
    """Roles vocabulary"""

    interface = IRole

    def __init__(self, *args, **kwargs):
        request = check_request()
        registry = request.registry
        translate = request.localizer.translate
        terms = [SimpleTerm(r.id, title=translate(r.title))
                 for n, r in registry.getUtilitiesFor(self.interface)]
        terms.sort(key=lambda x: x.title)
        super(RolesVocabulary, self).__init__(terms)

getVocabularyRegistry().register('PyAMS roles', RolesVocabulary)
