#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'

# import standard library
import logging
logger = logging.getLogger('PyAMS (security)')

# import interfaces
from pyams_security.interfaces import IProtectedObject, IRole, IPrincipalInfo, GrantedRoleEvent, RevokedRoleEvent, \
    IDefaultProtectionPolicy, IRoleProtectedObject
from zope.annotation.interfaces import IAnnotations

# import packages
from persistent import Persistent
from persistent.dict import PersistentDict
from pyams_utils.adapter import adapter_config
from pyams_utils.property import request_property
from pyams_utils.registry import query_utility
from pyramid.location import lineage
from pyramid.security import DENY_ALL, Everyone, Allow, ALL_PERMISSIONS, Authenticated
from pyramid.threadlocal import get_current_registry
from zope.container.contained import Contained
from zope.interface import implementer
from zope.lifecycleevent import ObjectCreatedEvent
from zope.location.location import locate
from zope.schema.fieldproperty import FieldProperty


@implementer(IRoleProtectedObject)
class RoleProtectedObject(Persistent, Contained):
    """Base class for object protected by roles"""

    inherit_parent_security = FieldProperty(IRoleProtectedObject['inherit_parent_security'])
    _everyone_permissions = FieldProperty(IRoleProtectedObject['everyone_permissions'])
    _authenticated_permissions = FieldProperty(IRoleProtectedObject['authenticated_permissions'])
    inherit_parent_roles = FieldProperty(IRoleProtectedObject['inherit_parent_roles'])

    def __init__(self):
        self._principals_by_role = PersistentDict()
        self._roles_by_principal = PersistentDict()

    @property
    def everyone_permissions(self):
        permissions = self._everyone_permissions
        if (not permissions) and self.inherit_parent_security:
            for parent in lineage(self):
                if parent in (self, self.__parent__):
                    continue
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    permissions = protection.everyone_permissions
                if permissions:
                    break
        return permissions

    @everyone_permissions.setter
    def everyone_permissions(self, value):
        self._everyone_permissions = value

    @property
    def authenticated_permissions(self):
        permissions = self._authenticated_permissions
        if (not permissions) and self.inherit_parent_security:
            for parent in lineage(self):
                if parent in (self, self.__parent__):
                    continue
                protection = IProtectedObject(parent, None)
                if protection is not None:
                    permissions = protection.authenticated_permissions
                if permissions:
                    break
        return permissions

    @authenticated_permissions.setter
    def authenticated_permissions(self, value):
        self._authenticated_permissions = value

    def grant_role(self, role_id, principal_ids):
        registry = get_current_registry()
        if IRole.providedBy(role_id):
            role_id = role_id.id
        if isinstance(principal_ids, str):
            principal_ids = {principal_ids}
        role_principals = self._principals_by_role.get(role_id) or set()
        for principal_id in principal_ids:
            if IPrincipalInfo.providedBy(principal_id):
                principal_id = principal_id.id
            if principal_id not in role_principals:
                principal_roles = self._roles_by_principal.get(principal_id) or set()
                role_principals.add(principal_id)
                principal_roles.add(role_id)
                self._roles_by_principal[principal_id] = principal_roles
                self._principals_by_role[role_id] = role_principals
                registry.notify(GrantedRoleEvent(self, role_id, principal_id))

    def revoke_role(self, role_id, principal_ids):
        registry = get_current_registry()
        if IRole.providedBy(role_id):
            role_id = role_id.id
        if isinstance(principal_ids, str):
            principal_ids = {principal_ids}
        role_principals = self._principals_by_role.get(role_id) or set()
        for principal_id in principal_ids.copy():
            if IPrincipalInfo.providedBy(principal_id):
                principal_id = principal_id.id
            if principal_id in role_principals:
                principal_roles = self._roles_by_principal.get(principal_id) or set()
                role_principals.remove(principal_id)
                principal_roles.remove(role_id)
                if principal_roles:
                    self._roles_by_principal[principal_id] = principal_roles
                else:
                    del self._roles_by_principal[principal_id]
                if role_principals:
                    self._principals_by_role[role_id] = role_principals
                else:
                    del self._principals_by_role[role_id]
                registry.notify(RevokedRoleEvent(self, role_id, principal_id))

    def get_principals(self, role_id):
        if IRole.providedBy(role_id):
            role_id = role_id.id
        return self._principals_by_role.get(role_id) or set()

    def get_roles(self, principal_id):
        if IPrincipalInfo.providedBy(principal_id):
            principal_id = principal_id.id
        return self._roles_by_principal.get(principal_id) or set()

    def get_permissions(self, principal_id):
        registry = get_current_registry()
        result = set()
        for role_id in self.get_roles(principal_id):
            role = registry.queryUtility(IRole, role_id)
            result |= role.permissions or set()
        return result

    @request_property(key=None)
    def __acl__(self):
        # always grant all permissions to system manager
        result = [(Allow, 'system:admin', ALL_PERMISSIONS)]
        # grant permission to everyone and authenticated
        if self.everyone_permissions:
            result.append((Allow, Everyone, self.everyone_permissions))
        if self.authenticated_permissions:
            result.append((Allow, Authenticated, self.authenticated_permissions))
        # grant access to all roles permissions
        for role_id in self._principals_by_role.keys():
            role = query_utility(IRole, role_id)
            if role is not None:
                result.append((Allow, 'role:{0}'.format(role_id), role.permissions))
        # stop inheritance if required
        if not self.inherit_parent_security:
            result.append(DENY_ALL)
        logger.debug('ACL = {0}'.format(str(result)))
        return result


ROLES_ANNOTATIONS_KEY = 'pyams_security.roles'


@adapter_config(context=IDefaultProtectionPolicy, provides=IRoleProtectedObject)
def ProtectedObjectFactory(context):
    """Default protected object factory"""
    annotations = IAnnotations(context)
    protection = annotations.get(ROLES_ANNOTATIONS_KEY)
    if protection is None:
        protection = annotations[ROLES_ANNOTATIONS_KEY] = RoleProtectedObject()
        get_current_registry().notify(ObjectCreatedEvent(protection))
        locate(protection, context)
    return protection


class ProtectedObject(object):
    """Base protected object class"""

    def __acl__(self):
        protected = IProtectedObject(self, None)
        if protected is not None:
            acl = protected.__acl__()
            if callable(acl):
                acl = acl(protected)
            return acl
        return []
